(*Gridlayout*)

unit udb_layout;

(*
  Teil des Programmes LZ-Dent, Abrechnungsprogramm für Dentallabore

  Copyright (C) 2015...  Dirk Graf       dirk@lz-dent.de / www.lz-dent.de

  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License as published by the Free
  Software Foundation; either version 2 of the License, or (at your option)
  any later version.

  This program is distributed in the hope that it will be useful, but WITHOUT
  ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
  FITNESS FOR A PARTICULAR PURPOSE. See the GNU General Public License for
  more details.

  You should have received a copy of the GNU General Public License along
  with this program; if not, write to the Free Software Foundation, Inc., 51
  Franklin St, Fifth Floor, Boston, MA 02110, USA

  ----------------------------------------------------------------------------

  Part of, LZ-Dent Abrechnungsprogramm für Dentallabore

  Copyright (C) 2015...  Dirk Graf       dirk@lz-dent.de / www.lz-dent.de

  Dieses Programm ist freie Software. Sie können es unter den Bedingungen der
  GNU General Public License, wie von der Free Software Foundation
  veröffentlicht, weitergeben und/oder modifizieren, entweder gemäß Version 2
  der Lizenz oder (nach Ihrer Option) jeder späteren Version.

  Die Veröffentlichung dieses Programms erfolgt in der Hoffnung, daß es Ihnen
  von Nutzen sein wird, aber OHNE IRGENDEINE GARANTIE, sogar ohne die
  implizite Garantie der MARKTREIFE oder der VERWENDBARKEIT FÜR EINEN
  BESTIMMTEN ZWECK. Details finden Sie in der GNU General Public License.

  Sie sollten ein Exemplar der GNU General Public License zusammen mit diesem
  Programm erhalten haben. Falls nicht, schreiben Sie an die Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110, USA.
*)


{$mode objfpc}{$H+}

interface

uses
  Classes, Controls,StdCtrls, SysUtils, ExtCtrls,  LMessages,
  Graphics,Menus;

type
  TGridPanel = class;
  TCustomGridPanel = class;
  TGridPanelControl = class;
  TGridPanelControlList = class;


  { TGridPanelControl }

  TGridPanelControl = class(TCollectionItem)
  private
    FActive: boolean;
    FCol: integer;
    FColSpan: integer;
    FControl: TControl;
    FRow: integer;
    FRowSpan: integer;
    procedure SetActive(AValue: boolean);
    procedure SetCol(AValue: integer);
    procedure SetColSpan(AValue: integer);
    procedure SetControl(const aControl: TControl);
    procedure SetRow(AValue: integer);
    procedure SetRowSpan(AValue: integer);
  protected
    procedure SetIndex(Value: integer); override;
    procedure AssignTo(Dest: TPersistent); override;
    function GPCollection: TGridPanelControlList;
    function GPOwner: TCustomGridPanel;
    property Index;
    constructor Create(ACollection: TCollection); override;
    function GetDisplayName: string; override;
  published
    (*The control itself *)
    property Control: TControl read FControl write SetControl;
    (*The column in the grid in which the control is displayed.*)
    property Col: integer read FCol write SetCol;
    (*The row in the grid in which the control is displayed.*)
    property Row: integer read FRow write SetRow;
    (*Width of control in columns*)
    property ColSpan: integer read FColSpan write SetColSpan;
    (*Height of control in rows*)
    property RowSpan: integer read FRowSpan write SetRowSpan;
    (*If not active, only the position is set, but not the height or width.*)
    property Active: boolean read FActive write SetActive;
  end;

  { TGridPanelControlList }

  TGridPanelControlList = class(TOwnedCollection)
  private
    function GetItem(Index: integer): TGridPanelControl;
    procedure SetItem(Index: integer; const AItem: TGridPanelControl);
  protected
    function FPOwner: TCustomGridPanel;

    function Add: TGridPanelControl;
    function AddControl(AControl: TControl; AIndex: integer = -1): TGridPanelControl;
    procedure RemoveControl(AControl: TControl);
  public
    constructor Create(AOwner: TPersistent);
    function ItemByDisplayName(aName:string):TGridPanelControl;
    Procedure DoSort;
  public
    function IndexOf(AControl: TControl): integer;
    property Items[Index: integer]: TGridPanelControl read GetItem write SetItem;
      default;
  end;

  { TCustomGridPanel }

  TCustomGridPanel = class(TCustomPanel)
  private
    FColCount: integer;
    FControlList: TGridPanelControlList;
    FDefaultRowHeight: Integer;
    FMarginLeft: integer;
    FMarginTop: integer;
    procedure SetColCount(AValue: integer);
    procedure SetControlList(const AControlList: TGridPanelControlList);
    procedure SetDefaultRowHeight(AValue: Integer);
    procedure SetMarginLeft(AValue: integer);
    procedure SetMarginTop(AValue: integer);
  protected
    inAdjust: boolean;
    procedure CMControlChange(var Message: TCMControlChange); message CM_CONTROLCHANGE;
    procedure AdJust;
    procedure SortItemList;

    function GetZeilenHoehe: integer; virtual;
    (*
    Rearranges all components
    *)
    procedure Resize; override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
    (*
    If a panel is subsequently made into a grid layout, all components
    can be re-registered at once.
    *)
    procedure Re_registers_all_child_components;
    (*
    Set the popupmenu for all childcontrols ans itself
    *)
    Procedure SetPopupMenu(aPop:TPopupMenu);
  public
    function GetControlIndex(AControl: TControl): integer;
    procedure SetControlIndex(AControl: TControl; Index: integer);
    (*
    The registred components
    *)
    property ControlList: TGridPanelControlList read FControlList write SetControlList;
    (*
    Distance of the components from the left and right edges and between the components.
    *)
    property MarginLeft: integer read FMarginLeft write SetMarginLeft default 4;
    (*
    Distance of the components from the top edge and between the components.
    *)
    property MarginTop: integer read FMarginTop write SetMarginTop default 4;
    (*
    Number of registered components
    *)
    property ColCount: integer read FColCount write SetColCount default 2;
    (*
    The row height in which the arrangement of the components is calculated.
    Does not include the distances between the components.
    *)
    property DefaultRowHeight:Integer read FDefaultRowHeight write SetDefaultRowHeight default 24;
  end;

  TGridPanel = class(TCustomGridPanel)
  published
    property Align;
    property Alignment;
    property Anchors;
    property AutoSize;
    property BevelInner;
    property BevelOuter default bvNone;
    property BevelWidth;
    property BiDiMode;
    property BorderWidth;
    property BorderSpacing;
    property BorderStyle;
    property Caption;
    property ColCount;
    property Color;
    property Constraints;
    property ControlList;
    property UseDockManager default True;
    property DockSite;
    property DoubleBuffered;
    property DefaultRowHeight;
    property DragCursor;
    property DragKind;
    property DragMode;
    property Enabled;


    property FullRepaint;
    property Font;
    property MarginLeft;
    property MarginTop;

    property ParentBiDiMode;
    property ParentColor;
    property ParentDoubleBuffered;
    property ParentFont;
    property ParentShowHint;
    property PopupMenu;
    property ShowHint;
    property TabOrder;
    property TabStop;
    property Visible;
    property OnAlignInsertBefore;
    property OnAlignPosition;
    property OnClick;
    property OnConstrainedResize;
    property OnContextPopup;
    property OnDockDrop;
    property OnDockOver;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDock;
    property OnEndDrag;
    property OnEnter;
    property OnExit;
    property OnGetSiteInfo;
    property OnMouseDown;
    property OnMouseEnter;
    property OnMouseLeave;
    property OnMouseMove;
    property OnMouseUp;
    property OnResize;
    property OnStartDock;
    property OnStartDrag;
    property OnUnDock;
  end;

implementation

uses Math;

{ TCustomGridPanel }

procedure TCustomGridPanel.SetControlList(const AControlList: TGridPanelControlList);
begin
  FControlList.Assign(AControlList);
end;

procedure TCustomGridPanel.SetDefaultRowHeight(AValue: Integer);
begin
  if FDefaultRowHeight=AValue then Exit;
  if FDefaultRowHeight>16 then
     FDefaultRowHeight:=AValue;
  AdJust;
end;

procedure TCustomGridPanel.SetColCount(AValue: integer);
begin
  if FColCount = AValue then
    Exit;
  FColCount := AValue;
  AdJust;
end;

procedure TCustomGridPanel.SetMarginLeft(AValue: integer);
begin
  if FMarginLeft = AValue then
    Exit;
  FMarginLeft := AValue;
  AdJust;
  SortItemList;
end;

procedure TCustomGridPanel.SetMarginTop(AValue: integer);
begin
  if FMarginTop = AValue then
    Exit;
  FMarginTop := AValue;
  AdJust;
end;

function GridPanelControlCompare(Item1, Item2: TCollectionItem): Integer;
Begin
  Result:=CompareStr(
  TGridPanelControl(Item1).DisplayName ,
  TGridPanelControl(Item2).DisplayName  );
end;

procedure TCustomGridPanel.CMControlChange(var Message: TCMControlChange);
var
  i, c, r: integer;
  w, h: integer;
  tc: TGridPanelControl;
  cnt: TControl;
begin
  if (csLoading in ComponentState) then
    Exit;
  if Message.Inserting and (Message.Control.Parent = Self) then
  begin
    cnt := Message.Control;
    DisableAlign;
    try
      cnt.Anchors := [akleft, aktop];
      cnt.AnchorSide[akTop].Control:=NIL;
      cnt.AnchorSide[akLeft].Control:=NIL;
      cnt.AnchorSide[akBottom].Control:=NIL;
      cnt.AnchorSide[akRight].Control:=NIL;

      cnt.Align := alNone;
      cnt.AutoSize := False;

      if cnt is TLabel then
        TLabel(cnt).Layout := tlCenter;
      if cnt is TComboBox then
        FDefaultRowHeight := cnt.Height;

      w := (Parent.ClientWidth - FMarginLeft) div FColCount;
      h := GetZeilenHoehe;


      tc := TGridPanelControl(FControlList.AddControl(cnt));

      tc.Col := (cnt.left + FMarginLeft) div w;
      Tc.Row := (cnt.Top + FMarginTop) div h;

      TC.ColSpan := succ(cnt.Width div w);
      if Tc.ColSpan < 1 then
        Tc.ColSpan := 1;
      Tc.RowSpan := succ(cnt.Height div h);
      if Tc.RowSpan < 1 then
        Tc.RowSpan := 1;
      AdJust;
    finally
      EnableAlign;
    end;
  end
  else
  begin

    FControlList.RemoveControl(Message.Control);
  end;
  SortItemList;
end;

procedure TCustomGridPanel.AdJust;
var
  itm: TGridPanelControl;
  xControl: TControl;
  xNewBounds: array of TRect;
  Spaltenbreite, ZeilenHoehe: integer;
  ll, ww, hh, tt, i: integer;


  function CompareRect(r1, r2: TRect): boolean;
  begin
    Result := (r1.Left = r2.Left) and (r1.Top = r2.Top) and
      (r1.Width = r2.Width) and (r1.Height = r2.Height);
  end;

begin
  if csLoading in ComponentState then
    exit;
  if inAdjust then
    exit;
  if FColCount = 0 then
    exit;

  try
    inAdjust := True;

    Spaltenbreite := (ClientRect.Width - FMarginLeft) div FColCount;
    ZeilenHoehe := GetZeilenHoehe;

    for i := 0 to FControlList.Count - 1 do
    begin
      itm := TGridPanelControl(FControlList.Items[i]);
        xControl := itm.Control;

        ll := SpaltenBreite * itm.col + FMarginLeft;
        ww := (SpaltenBreite * itm.ColSpan) - FMarginLeft;
        tt := ZeilenHoehe * itm.Row + FMarginTop;
        hh := ZeilenHoehe * Itm.RowSpan - FMarginTop;
        try
        if itm.Active then
        begin
          xControl.SetBounds(ll, tt, ww, hh);
        end else
          xControl.SetBounds(ll, tt, xControl.Width, xControl.Height);
        except
        end;
    end;
  finally
    inAdjust := False;
  end;
end;

procedure TCustomGridPanel.SortItemList;
begin
  FControlList.Sort(  @GridPanelControlCompare );
end;

function TCustomGridPanel.GetZeilenHoehe: integer;
var
  C: TComboBox;
begin

  c:=TComboBox.create(self);
  try
  //  c.Parent:=self;
    c.Text:='Pg';
    Result := c.Height + FMarginTop;
  finally
    c.free;
  end;
//  Result := FDefaultRowHeight + FMarginTop;
end;

procedure TCustomGridPanel.Resize;
begin
  inherited Resize;
  // if ComponentState * [csLoading, csUpdating,  csDestroying] = [] then
  if inAdjust then exit;
  AdJust;
end;

constructor TCustomGridPanel.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  inAdjust := False;
  FControlList := TGridPanelControlList.Create(Self);
  FColCount := 2;
  FMarginLeft := 4;
  FMarginTop := 4;
  FDefaultRowHeight:=24;
  BevelOuter:=bvNone;
  Color:=clSkyBlue;

end;

destructor TCustomGridPanel.Destroy;
begin
  FControlList.Free;
  inherited Destroy;
end;

procedure TCustomGridPanel.Re_registers_all_child_components;
var
  i:integer;
  cnt:TControl;
  w,h:Integer;
  tc:TGridPanelControl;
begin
  FControlList.Clear;
  DisableAlign;
  try
  for i:=0 to ControlCount-1 do
  Begin
    cnt := Controls[i];

      cnt.Anchors := [akleft, aktop];
      cnt.Align := alNone;
      cnt.AutoSize := False;

      if cnt is TLabel then
        TLabel(cnt).Layout := tlCenter;
      if cnt is TCustomComboBox then
        FDefaultRowHeight := cnt.Height;

      w := (Parent.ClientWidth - FMarginLeft) div FColCount;
      h := GetZeilenHoehe;


      tc := TGridPanelControl(FControlList.AddControl(cnt));

      tc.Col := (cnt.left + FMarginLeft) div w;
      Tc.Row := (cnt.Top + FMarginTop) div h;

      TC.ColSpan := succ(cnt.Width div w);
      if Tc.ColSpan < 1 then
        Tc.ColSpan := 1;
      Tc.RowSpan := succ(cnt.Height div h);
      if Tc.RowSpan < 1 then
        Tc.RowSpan := 1;
      end;
          AdJust;
    finally
      EnableAlign;
    end;
end;

procedure TCustomGridPanel.SetPopupMenu(aPop: TPopupMenu);
var i:integer;
begin
  PopupMenu:=aPop;
  for i:=0 to FControlList.Count-1 do
    TGridPanelControl(FControlList.Items[i]).Control.PopupMenu:=aPop ;

end;

function TCustomGridPanel.GetControlIndex(AControl: TControl): integer;
begin
  Result := FControlList.IndexOf(AControl);
end;

procedure TCustomGridPanel.SetControlIndex(AControl: TControl; Index: integer);
var
  CurIndex: integer;
begin
  CurIndex := GetControlIndex(AControl);
  if (CurIndex > -1) and (CurIndex <> Index) and (Index < FControlList.Count) then
  begin
    FControlList.Items[CurIndex].Index := Index;
    Realign;
  end;
end;

{ TGridPanelControlList }

function TGridPanelControlList.GetItem(Index: integer): TGridPanelControl;
begin
  Result := TGridPanelControl(inherited GetItem(Index));
end;

procedure TGridPanelControlList.SetItem(Index: integer; const AItem: TGridPanelControl);
begin
  inherited SetItem(Index, AItem);
end;

function TGridPanelControlList.FPOwner: TCustomGridPanel;
begin
  Result := TCustomGridPanel(GetOwner);
end;

function TGridPanelControlList.Add: TGridPanelControl;
begin
  Result := TGridPanelControl(inherited Add);
end;

function TGridPanelControlList.AddControl(AControl: TControl;
  AIndex: integer): TGridPanelControl;
var
  I: integer;
begin
  if IndexOf(AControl) >= 0 then
    Exit;

  if AIndex = -1 then
    for I := 0 to Count - 1 do
      if not Assigned(Items[I].Control) then
      begin
        AIndex := I;
        break;
      end;

  if AIndex = -1 then
    Result := Add
  else
    Result := Items[AIndex];
  Result.FControl := AControl;

end;

procedure TGridPanelControlList.RemoveControl(AControl: TControl);
var
  I: integer;
begin
  for I := Count - 1 downto 0 do
    if Items[I].Control = AControl then
    begin
      Items[I].FControl := nil;
      Delete(I);
      Exit;
    end;
end;

constructor TGridPanelControlList.Create(AOwner: TPersistent);
begin
  Assert(AOwner is TCustomGridPanel);
  inherited Create(AOwner, TGridPanelControl);
end;

function TGridPanelControlList.ItemByDisplayName(aName: string
  ): TGridPanelControl;
var i:integer;
begin
  Result:=NIL;
  for i:=0 to count-1 do
  if SameText( TGridPanelControl(Items[i] ).DisplayName,aName) then
  begin
    Result:=TGridPanelControl(Items[i] );
    exit;
  end;
end;

procedure TGridPanelControlList.DoSort;
var
  ts:TStringList;
  tgc:TGridPanelControl;
  i:integer;
begin
  BeginUpdate;
  ts:=TStringList.Create;
  try
    for i:=0 to Count-1 do
      ts.add(Items[i].DisplayName);
    ts.sort;
    for i:=0 to ts.Count-1 do
    Begin
       tgc:=TGridPanelControl(ItemByDisplayName( ts[i]));
       if Assigned(tgc) then
         tgc.SetIndex(i);
    end;
  finally
    ts.Free;
  end;
  EndUpdate;
end;

function TGridPanelControlList.IndexOf(AControl: TControl): integer;
begin
  for Result := 0 to Count - 1 do
    if Items[Result].Control = AControl then
      Exit;
  Result := -1;

end;

{ TGridPanelControl }

procedure TGridPanelControl.SetCol(AValue: integer);
begin
  if FCol = AValue then
    Exit;
  FCol := AValue;
  if Assigned(GPOwner) then
    GPOwner.AdJust;
end;

procedure TGridPanelControl.SetActive(AValue: boolean);
begin
  if FActive = AValue then
    Exit;
  FActive := AValue;
end;

procedure TGridPanelControl.SetColSpan(AValue: integer);
begin
  if FColSpan = AValue then
    Exit;
  FColSpan := AValue;
  if Assigned(GPOwner) then
    GPOwner.AdJust;
end;

procedure TGridPanelControl.SetControl(const aControl: TControl);
begin
  if FControl = aControl then
    Exit;
  Assert(FControl = nil);
  FControl := aControl;
end;

procedure TGridPanelControl.SetRow(AValue: integer);
begin
  if FRow = AValue then
    Exit;
  FRow := AValue;
  if Assigned(GPOwner) then
    GPOwner.AdJust;
end;

procedure TGridPanelControl.SetRowSpan(AValue: integer);
begin
  if FRowSpan = AValue then
    Exit;
  FRowSpan := AValue;
  if Assigned(GPOwner) then
    GPOwner.AdJust;
end;

procedure TGridPanelControl.SetIndex(Value: integer);
begin
  inherited SetIndex(Value);
  if GPOwner.ComponentState * [csLoading, csUpdating, csDestroying] = [] then
    GPOwner.ReAlign;
end;

procedure TGridPanelControl.AssignTo(Dest: TPersistent);
var
  xDest: TGridPanelControl;
begin
  if Dest is TGridPanelControl then
  begin
    xDest := TGridPanelControl(Dest);
    xDest.Col := Self.Fcol;
    xDest.Row := Self.FRow;
  end
  else
    inherited AssignTo(Dest);
end;

function TGridPanelControl.GPCollection: TGridPanelControlList;
begin
  Result := Collection as TGridPanelControlList;
end;

function TGridPanelControl.GPOwner: TCustomGridPanel;
begin
  Result := GPCollection.FPOwner;
end;

constructor TGridPanelControl.Create(ACollection: TCollection);
begin
  inherited Create(ACollection);
  FActive := True;
end;

function TGridPanelControl.GetDisplayName: string;
begin
  if Assigned(FControl) then
    Result:=FControl.Name+' ('+FControl.ClassName+') '
  else
    Result:=inherited GetDisplayName;
end;


end.
